/*
 * \file Server.cpp
 *
 *  \date Oct 16, 2013
 *  \author Alexandre Tuleu
 */

#include "Server.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <cstring>
#include <errno.h>

#include <sstream>


namespace net {

Server::Server(Type type)
	: d_type(type)
	, d_fd(-1)
	, d_maxConnections(10){
	if(d_type != TCP && d_type != UDP) {
		throw std::logic_error("Unsupported server type");
	}
}


Server::~Server(){
	Clean();
}

void Server::Clean() {
	if(d_fd == -1) {
		return;
	}
	close(d_fd);
	d_fd = -1;
}

void Server::SetMaxPendingConnections(unsigned int max){
	d_maxConnections = max;
}

unsigned Server::MaxPendingConnections() const{
	return d_maxConnections;
}

void Server::Listen(const std::string & address ,
                    unsigned int port){

	addrinfo hints, *servinfo;
	memset(&hints,0,sizeof(addrinfo));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = (d_type == TCP ? SOCK_STREAM : SOCK_DGRAM);
	const char * addr;
	if(address.empty()) {
		hints.ai_flags = AI_PASSIVE;
		addr = NULL;
	} else {
		addr = address.c_str();
	}

	std::ostringstream portStr;
	portStr << port;

	int res = getaddrinfo(addr,portStr.str().c_str(),&hints,&servinfo);

	if( res ) {
		throw std::runtime_error("Could not resolve '"
		                         + (address.empty() ?  std::string("<any_address>")
		                                            : address)
		                         + ":" + portStr.str() + "' : "
		                         + gai_strerror(res));
	}

	std::ostringstream errors;
	for(addrinfo * p = servinfo; p != NULL; p = p->ai_next) {
		d_fd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
		if ( d_fd == -1 ) {
			errors << "* Could not create socket for host '" << p->ai_canonname
			       << "' : " << strerror(errno) << "." << std::endl;
			continue;
		}
		int yes = 1;
		res = setsockopt(d_fd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int));
		if(res == -1) {
			Clean();
			errors << "* Could not set socket option : "
			       << strerror(errno) << "." << std::endl;
			continue;
		}

		res = bind(d_fd,p->ai_addr, p->ai_addrlen);
		if(res == -1) {
			Clean();
			errors << "* Could not bind socket : "
			       << strerror(errno) << "."
			       << std::endl;
			continue;
		}

		break;

	}
	freeaddrinfo(servinfo);
	if (d_fd == -1) {
		throw std::runtime_error("Could not bind to '"
		                          + (address.empty() ?  std::string("<any_address>")
		                                             : address)
		                          + ":" + portStr.str() + "', got error(s):\n"
		                          + errors.str() );
	}

	res = listen(d_fd,(int)d_maxConnections);
	if( res == -1 ){
		throw std::runtime_error("Could not listen on binded socket : " + std::string(strerror(errno)) + ".");
	}
}


class SocketFromServer : public Socket {
public :
	SocketFromServer(int fd) { SetFd(fd); };
	virtual ~SocketFromServer() {};
};

Socket::Ptr Server::Accept(){
	struct sockaddr_storage theirAddr;
	socklen_t sinSize(sizeof(theirAddr));
	int newfd = accept(d_fd,(struct sockaddr *)&theirAddr,&sinSize);
	if( newfd == -1) {
		throw std::runtime_error("Could not accept incoming connections : " + std::string(strerror(errno)) + ".");
	}

	return Socket::Ptr(new SocketFromServer(newfd));
}

} /* namespace net */
