/*
 * \file Socket.cpp
 *
 *  \date Sep 18, 2013
 *  \author Alexandre Tuleu
 */

#include "Socket.h"

#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <netdb.h>

#include <string.h>

#include <biorob-cpp/os/SystemError.h>

#include <sstream>

namespace net {

SocketFullIOError::SocketFullIOError(IOType type,
                                     size_t total,
                                     size_t performed,
                                     unsigned int retries)
	: std::runtime_error(Stringify(type,total,performed,retries))
	, d_type(type)
	, d_total(total)
	, d_performed(performed)
	, d_retries(retries){
}

SocketFullIOError::~SocketFullIOError() throw(){}

SocketFullIOError::IOType SocketFullIOError::Type() const {
	return d_type;
}

size_t SocketFullIOError::TotalSize() const{
	return d_total;
}

size_t SocketFullIOError::PerformedSize() const{
	return d_performed;
}

size_t SocketFullIOError::Retries() const{
	return d_retries;
}

std::string SocketFullIOError::Stringify(IOType type,
                                         size_t total ,
                                         size_t performed ,
                                         unsigned int retries){
	std::ostringstream os;
	std::string action;
	switch(type) {
	case READ :
		action = "read";
		os << "ReadAll()";
		break;
	case WRITE :
		action = "written";
		os << "WriteAll()";
		break;
	default:
		action = "cooked";
		os << "unknown";
		break;
	}
	os << " of " << total << " bytes failed after " << retries
	   << ", only " << performed << " bytes " << action << ".";
	return os.str();
}



Socket::Socket()
	: d_fd(-1)
	, d_recvFlags(0)
	, d_sendFlags(0)
	, d_maxRetries(20){
}

Socket::Socket(int fd)
	: d_fd(fd)
	, d_recvFlags(0)
	, d_sendFlags(0)
	, d_maxRetries(20){
}

Socket::~Socket(){
	Clean();
}

size_t Socket::Read(uint8_t* buffer, size_t length){
	int byteRead = recv(d_fd,buffer,length,d_recvFlags);
	if (byteRead < 0 ){
		throw os::SystemError("recv() returned error ",errno);
	}
	return (size_t)byteRead;
}

size_t Socket::Write(uint8_t* buffer , size_t length){
	int byteWritten = send(d_fd,buffer,length,d_sendFlags);
	if(byteWritten < 0) {
		throw os::SystemError("send() returned error ",errno);
	}
	return (size_t)byteWritten;
}

void Socket::ReadAll(uint8_t* buffer , size_t length){
	unsigned int trials = 0;
	size_t byteRead = 0;
	do {
		byteRead += Read(buffer + byteRead,length - byteRead);
		++trials;
	}while(trials < d_maxRetries && byteRead < length);

	if(byteRead < length) {
		throw SocketFullIOError(SocketFullIOError::READ,
		                        length,
		                        byteRead,
		                        d_maxRetries);
	}
}

void Socket::WriteAll(uint8_t* buffer , size_t length){
	unsigned int trials = 0;
	size_t byteWritten = 0;
	do {
		byteWritten += Write(buffer + byteWritten, length - byteWritten);
		++trials;
	}while(trials < d_maxRetries && byteWritten < length);

	if(byteWritten < length) {
		throw SocketFullIOError(SocketFullIOError::WRITE,
		                        length,
		                        byteWritten,
		                        d_maxRetries);
	}
}

void Socket::SetNonBlocking(bool block){
	int flags = fcntl(d_fd,F_GETFL,0);
	if ( flags < 0 ) {
		throw os::SystemError("fcntl() returned error ",errno);
	}


	if(block) {
		p_call(fcntl,d_fd,F_SETFL,O_NONBLOCK | flags);
	} else {
		p_call(fcntl,d_fd,F_SETFL, flags & (~O_NONBLOCK));
	}
}

void Socket::SetFd(int fd){
	Clean();
	d_fd = fd;
}

int Socket::GetSocketType(int fd) {
	//check if this is a socket
	sockaddr addr;
	socklen_t length = sizeof(sockaddr);
	try {
		p_call(getsockname,fd,&addr,&length);
	} catch (const os::SystemError & e) {
		if (e.Error() == ENOTSOCK ) {
			throw std::logic_error("The file descriptor is not a POSIX Socket.");
		}
		//we throw anyway in case of error
		throw;
	}

	int socketType;
	length = sizeof(int);
	p_call(getsockopt,fd,SOL_SOCKET,SO_TYPE,&socketType,&length);
	return socketType;
}



int Socket::ConnectSocket(const std::string & host,
                          unsigned int port,
                          int type) {
	addrinfo hints, *servinfo;
	memset(&hints,0,sizeof(addrinfo));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = type;

	std::ostringstream portStr;
	portStr << port;

	int res = getaddrinfo(host.c_str(), portStr.str().c_str(),&hints,&servinfo);
	if( res ) {
		throw std::runtime_error("Could not resolve '" + host + ":"
		                         + portStr.str() + "' : " + gai_strerror(res));
	}

	int socket_fd = -1;

	std::ostringstream errors;
	for( addrinfo *p = servinfo; p != NULL ; p = p->ai_next) {

		socket_fd = socket(p->ai_family,p->ai_socktype,p->ai_protocol);
		if( socket_fd == -1 ){
			errors << "* Could not create socket for host '" << p->ai_canonname
			       << "' : " << strerror(errno) << "." << std::endl;
			continue;
		}
		res =  connect(socket_fd,p->ai_addr,p->ai_addrlen);
		if (res) {
			close(socket_fd);
			socket_fd = -1;
			errors << "* Could not connect socket for host '" << p->ai_canonname
			       << "' : " << strerror(errno) << std::endl;
			continue;
		}
	}

	if (socket_fd == -1 ) {
		throw std::runtime_error("Could not connect to '" + host + ":"
		                         + portStr.str() + "', got error(s):\n"
		                         + errors.str());
	}

	return socket_fd;
}

void Socket::SetReceiveFlags(int flags){
	d_recvFlags = flags;
}

void Socket::SetSendFlags(int flags){
	d_sendFlags = flags;
}

unsigned int Socket::GetMaxRetries(){
	return d_maxRetries;
}

void Socket::SetMaxRetries(unsigned int maxRetries){
	d_maxRetries = maxRetries;
}

void Socket::Clean(){
	if(d_fd != -1 ) {
		close(d_fd);
		d_fd = -1;
	}
}

} /* namespace net */
