/*
 * \file CliArgParser.cpp
 *
 *  \date Oct 10, 2013
 *  \author Alexandre Tuleu
 */

#include "CliArgParser.h"

#include <stdexcept>
#include <iostream>
#include <cstdlib>


namespace options {

CliArgParser::CliArgParser(const std::string& description)
	: d_helpEnabled(true)
	, d_ignoreUnusedArgs(false)
	, d_description(description) {
	if(d_description.empty()) {
		throw std::logic_error("Command Line Description is mandatory");
	}
}

void CliArgParser::AppendUsage(const std::string& usage) {
	d_usages.push_back(usage);
}

void CliArgParser::EnableHelp(bool on) {
	d_helpEnabled = on;
}

bool CliArgParser::HelpEnabled() const {
	return d_helpEnabled;
}

void CliArgParser::IgnoreUnusedArguments(bool on) {
	if(on) {
		throw std::logic_error("Not yet implemented.");
	}
	d_ignoreUnusedArgs = on;
}

bool CliArgParser::UnusedArgumentsIgnored() const {
	return d_ignoreUnusedArgs;
}

void CliArgParser::AddOption(const std::string& name,
                             bool& value,
                             const std::string& description,
                             char shortName) {
	CheckOptionName(name,shortName);

	ValueHolderBase::Ptr holder(new ValueHolder<bool>(name,
	                                                  description,
	                                                  shortName,
	                                                  false,
	                                                  value));
	d_longValues[name] = holder;
	if(shortName != 0) {
		d_shortValues[shortName] = holder;
	}
}

void CliArgParser::PrintHelp(std::ostream & out) const {

	out << d_description << std::endl;
	if(!d_usages.empty()) {
		out << "usage:" << std::endl;
		for(std::vector<std::string>::const_iterator u = d_usages.begin();
		    u != d_usages.end();
		    ++u) {
			out << "  " << *u << std::endl;
		}
	}

	if(d_longValues.empty() && d_helpEnabled == false) {
		return;
	}
	out << "arguments:" << std::endl;
	if(d_helpEnabled) {
		out << "  --help/-h : Prints this help message and exits." 
		    << std::endl;
	}

	for(ValuesByLong::const_iterator vh = d_longValues.begin();
	    vh != d_longValues.end();
	    ++vh) {
		out << "  ";
		vh->second->PrintDescription(out);
		out << std::endl;
	}
}

std::vector<std::string>
CliArgParser::Parse(int argc,
                    char ** argv,
                    ActionOnError actionOnError) {
	std::vector<std::string> result;
	switch(actionOnError) {
	case EXIT_ON_ERROR :
		ParseOrExit(argc,argv,result);
		break;
	case THROW_ON_ERROR :
		ParseOrThrow(argc,argv,result);
		break;
	default :
		throw std::logic_error("Wrong action or error passed to CliArgumentParser::Parse");
		break;
	}

	return result;
}


	void CliArgParser::ParseOrExit(int argc,
	                               char ** argv,
	                               std::vector<std::string> & values) {
	try {
		ParseOrThrow(argc,argv,values);
	} catch (const std::exception & e) {
		std::cerr << "Error during argument parsing : "
		          << e.what() << std::endl;
		PrintHelp(std::cerr);
		std::exit(1);
	}
}

	void CliArgParser::ParseOrThrow(int argc,
	                                char ** argv,
	                                std::vector<std::string> & values) {
	values.clear();
	if( argc < 1 || std::string(argv[0]).empty() ) {
		std::ostringstream os;
		os << "Bad command line argument number " << argc 
		   << " it should be at least be 1.";
		throw std::logic_error(os.str());
	}
	for(int i = 1; i < argc; ++i) {
		std::vector<std::string> inlinedValues,optionNames;
		ExtractOptionName(argv[i],optionNames,inlinedValues);

		if(optionNames.empty() ) {
			//this is not an option, we add it to the list of the
			//unparsed values.
			values.push_back(argv[i]);
			continue;
		}
		//syntax for multiple short
		if(optionNames.size() > 1) {
			for(std::vector<std::string>::size_type j = 0;
			    j < optionNames.size();
			    ++j) {
				if(inlinedValues.size() > j) {
					ParseShortOption(optionNames[j],
					                 inlinedValues[j]);
				} else {
					ParseShortOption(optionNames[j],"");
				}
			}
			continue;
		}

		ValueHolderBase::Ptr vh;
		if(optionNames[0].size() == 1) {
			if(optionNames[0][0] == 'h') {
				PrintHelpAndExitIfEnabled();
				continue;
			}

			ValuesByShort::const_iterator fi =
				d_shortValues.find(optionNames[0][0]);
			if (fi != d_shortValues.end()) {
				vh = fi->second;
			}
		} else {
			if(optionNames[0] == "help" ){
				PrintHelpAndExitIfEnabled();
				continue;
			}
			ValuesByLong::const_iterator fi = d_longValues.find(optionNames[0]);
			if (fi != d_longValues.end()) {
				vh = fi->second;
			}
		}

		if(!vh) {
			throw std::runtime_error(std::string("Unknown flag '")
			                                     + argv[i] + "'" );
		}

		if(!inlinedValues.empty()) {
			vh->ConsumeArg(inlinedValues[0].c_str());
			continue;
		}

		int consumed = 0;
		for(int j = 1 ;
		    (j + i < argc) && (j <= vh->MaxExpectedArgs());
		    ++j) {
			std::vector<std::string> nextOptionNames,nextInlinedValues;
			ExtractOptionName(argv[i + j],
			                  nextOptionNames,
			                  nextInlinedValues);
			if(!nextOptionNames.empty()) {
				break;
			}
			vh->ConsumeArg(argv[i + j]);
			++consumed;
		}
		if(consumed == 0) {
			vh->ConsumeArg("");
		}
		i += consumed;

	}

}

CliArgParser::ValueHolderBase::ValueHolderBase(const std::string & name,
                                               const std::string & description,
                                               char shortName,
                                               bool mandatory,
                                               int maxExpectedArgs)
	: d_name(name)
	, d_description(description)
	, d_shortName(shortName)
	, d_mandatory(mandatory)
	, d_modified(false)
	, d_maxExpectedArgs(maxExpectedArgs) {
	if(description.empty()) {
		throw std::logic_error("Suitable description required : this option is not empty, isn't it ?");
	}
}

CliArgParser::ValueHolderBase::~ValueHolderBase() {
}

const std::string & CliArgParser::ValueHolderBase::Name() const {
	return d_name;
}

const std::string & CliArgParser::ValueHolderBase::Description() const {
	return d_description;
}

std::string CliArgParser::ValueHolderBase::FullName() const {
	if ( d_shortName != 0 ) {
		std::ostringstream os;
		os << "--" << d_name << "/-" << d_shortName;
		return os.str();
	}
	return "--" + d_name;
}

bool CliArgParser::ValueHolderBase::CheckMandatory() const {
	return !d_mandatory || d_modified;
}

void CliArgParser::ValueHolderBase::MarkModified() {
	d_modified = true;
}

bool CliArgParser::ValueHolderBase::Modified() const {
	return d_modified;
}


int CliArgParser::ValueHolderBase::MaxExpectedArgs() const {
	return d_maxExpectedArgs;
}


void CliArgParser::CheckOptionName(const std::string & name,char shortName) const {
	CheckOptionNameCorrectness(name,shortName);
	CheckOptionNameAvailability(name,shortName);
}
void CliArgParser::CheckOptionNameCorrectness(const std::string & name,char shortName) {
#define throw_bad_long_name() do {	  \
	throw std::logic_error("Invalid option name '" + name + "'"); \
 } while(0)

	//not using regex, too complicated.
	// must start by a alphabetic character and be non-empty
	if(name.size() < 2) {
		throw_bad_long_name();
	}
	
	if(!std::isalpha(name[0])) {
		throw_bad_long_name();
	}
	
	//now we check that amny following char are in [:word:\-]
	for(std::string::size_type i = 1;
	    i < name.size();
	    ++i) {
		if(std::isalnum(name[i]) || name[i] == '_'  || name[i] == '-') {
			continue;
		}
		throw_bad_long_name();
	}
	//this is the case, so long name  is checked
	#undef throw_bad_long_name

	if(name == "help") {
		throw std::logic_error("'help' name is reserved for help message option");
	}

	if(shortName == 0) {
		return;
	}

	if( !(shortName >='a' && shortName <= 'z') &&
	    !(shortName >= 'A' && shortName <= 'Z') ) {
		std::ostringstream os;
		os << "invalid short option name '" << shortName << "'";
		throw std::logic_error(os.str());
	}

	if( shortName == 'h' ) {
		throw std::logic_error("'h' shortname is reserved for help message option");
	}

}
void CliArgParser::CheckOptionNameAvailability(const std::string & name,
                                               char shortName) const {
	ValuesByLong::const_iterator fi = d_longValues.find(name);
	if (fi != d_longValues.end()) {
		throw std::logic_error("option name '--" + name + "' is already in use");
	}

	if(shortName == 0) {
		return;
	}

	ValuesByShort::const_iterator fiC = d_shortValues.find(shortName);
	if( fiC != d_shortValues.end()) {
		std::ostringstream os;
		os << "Option short name '-" << shortName << "' is already in use";
		throw std::logic_error(os.str());
	}
}


void CliArgParser::ExtractOptionName(const char * arg,
                                     std::vector<std::string> & name,
                                     std::vector<std::string> & inlineValue) {

	//first we clear
	name.clear();
	inlineValue.clear();

	std::string argStr(arg);
	if(argStr.empty()) {
		throw std::logic_error("Empty argument '" + argStr + "'");
	}

	if(argStr[0] != '-') {
		return;
	}

	if(argStr.size() < 2) {
		return;
	}


	if(std::isdigit(argStr[1])) {
		return;
	}

	if(std::isalpha(argStr[1])) {
		for(std::string::size_type i = 1;
		    i < argStr.size();
		    ++i) {
			if(argStr[i] == '=') {
				if(i == 1)  {
					throw std::runtime_error("Invalid flag '" + argStr + "'");
				}
				std::string inlineValues = argStr.substr(i+1);
				inlineValue.assign(name.size(),inlineValues);
				break;
			}
			if(std::isalpha(argStr[i])) {
				name.push_back(argStr.substr(i,1));
				continue;
			}
			throw std::runtime_error("Invalid flag '" + argStr + "'");
		}
	}

	if(argStr[1] != '-') {
		return;
	}

	if(argStr.size() < 3) {
		return;
	}

	std::string::size_type pos = argStr.find("=");
	if(pos != std::string::npos ) {
		name.push_back(argStr.substr(2, pos - 2));
		inlineValue.push_back(argStr.substr(pos + 1));
	} else {
		name.push_back(argStr.substr(2));
	}
}

void CliArgParser::ParseShortOption(const std::string & name,
                                    const std::string & inlineValue) {
	if(name.size() != 1 ) {
		throw std::logic_error("Unexepected short name size.");
	}

	ValuesByShort::const_iterator fi = d_shortValues.find(name[0]);
	if(fi == d_shortValues.end()) {
		throw std::runtime_error("Unknown flag '-" + name + "'");
	}

	fi->second->ConsumeArg(inlineValue.c_str());

}

void CliArgParser::PrintHelpAndExitIfEnabled() const {
	if(d_helpEnabled == false) {
		return;
	}
	PrintHelp(std::cout);
	exit(0);
}

} /* namespace options */


