/*
 * \file CliArgParser.imp.h
 *
 *  \date Oct 10, 2013
 *  \author Alexandre Tuleu
 */

#ifndef BIOROB_CPP_CLIARGPARSER_IMP_H_
#define BIOROB_CPP_CLIARGPARSER_IMP_H_

#include "CliArgParser.h"

#include <sstream>
#include <iomanip>
#include <stdexcept>
#include <limits>

namespace options {

template <typename T>
inline CliArgParser::ValueHolder<T>::ValueHolder(const std::string & name,
                                                 const std::string & description,
                                                 char shortName,
                                                 bool mandatory,
                                                 T & value)
	: ValueHolderBase(name,description,shortName,mandatory,1)
	, d_value(value){
}

template <>
inline CliArgParser::ValueHolder<bool>::ValueHolder(const std::string & name,
                                                 const std::string & description,
                                                 char shortName,
                                                 bool mandatory,
                                                 bool & value)
	: ValueHolderBase(name,description,shortName,mandatory,0)
	, d_value(value){
	d_value = false;
}


template <typename T>
inline CliArgParser::ValueHolder<T>::~ValueHolder(){
}

template <typename T>
inline void CliArgParser::ValueHolder<T>::ConsumeArg(const char * arg) {
	if(Modified()) {
		throw std::runtime_error("Multiple instance of parameter '" + FullName() + "'.");
	}
	std::istringstream iss(arg);
	iss >> d_value;
	MarkModified();
}

template <>
inline void CliArgParser::ValueHolder<std::string>::ConsumeArg(const char * arg) {
	if(Modified()) {
		throw std::runtime_error("Multiple instance of parameter '" + FullName() + "'.");
	}
	d_value = std::string(arg);
	MarkModified();
}

template <>
inline void CliArgParser::ValueHolder<bool>::ConsumeArg(const char * arg) {
	std::string argStr(arg);
	if(!argStr.empty()) {
		throw std::runtime_error("option does not accept value");
	}
	d_value = true;
}

template <typename T>
inline void CliArgParser::ValueHolder<T>::PrintDescription(std::ostream & out) const {
	out << FullName() << " : " << Description() << " (default: " << d_value << ")";
}

template <>
inline void CliArgParser::ValueHolder<bool>::PrintDescription(std::ostream & out) const {
	out << FullName() << " : " << Description() << " (default: disabled)";
}


template <typename T>
inline CliArgParser::ListValueHolder<T>::ListValueHolder(const std::string & name,
                                                         const std::string & description,
                                                         char shortName,
                                                         std::vector<T> & values)
	: ValueHolderBase(name,description,shortName,false,std::numeric_limits<int>::max())
	, d_values(values){
}

template <typename T>
inline CliArgParser::ListValueHolder<T>::~ListValueHolder(){
}

template <typename T>
inline void CliArgParser::ListValueHolder<T>::ConsumeArg(const char * arg) {
	std::istringstream iss(arg);
	T value;
	iss >> value;
	d_values.push_back(value);
}

template <>
inline void CliArgParser::ListValueHolder<std::string>::ConsumeArg(const char * arg) {
	d_values.push_back(std::string(arg));
}

template <typename T>
inline void CliArgParser::ListValueHolder<T>::PrintDescription(std::ostream & out) const {
	out << FullName() << " : " << Description() << " (list of values)";
}

template <typename T>
inline void CliArgParser::AddParameter(const std::string & name,
                                       T & value,
                                       const std::string & description,
                                       bool mandatory,
                                       char shortName){
	CheckOptionName(name,shortName); //throws on error;

	ValueHolderBase::Ptr holder(new ValueHolder<T>(name,description,shortName,mandatory,value));
	d_longValues[name] = holder;
	if(shortName != 0) {
		d_shortValues[shortName] = holder;
	}

}

template <>
inline void CliArgParser::AddParameter(const std::string & /*name*/,
                                       bool & /*value*/,
                                       const std::string & /*description*/,
                                       bool /*mandatory*/,
                                       char /*shortName*/) {
	throw std::logic_error("Invalid type of booleean for a parameter, please use an option");
}

template <typename T>
inline void CliArgParser::AddListParameter(const std::string & name,
                                           std::vector<T> & value,
                                           const std::string & description,
                                           char shortName){
	CheckOptionName(name,shortName);
	ValueHolderBase::Ptr holder(new ListValueHolder<T>(name,description,shortName,value));
	d_longValues[name] = holder;
	if(shortName != 0) {
		d_shortValues[shortName] = holder;
	}
}


} /*namespace options */



#endif /* BIOROB_CPP_CLIARGPARSER_IMP_H_ */
