/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \file Option.cpp
 *
 *  \date Nov 28, 2011
 *  \author Alexandre Tuleu
 */

#include "Option.h"

using namespace options;


SectionAndOptionBase::SectionAndOptionBase(const std::string & name,
                                           const std::string & comment)
	: d_name(name)
	, d_comment(comment)
	, d_short(0)
	, d_isUnknown(false){
	SetNameFromString(name);
}



SectionAndOptionBase::~SectionAndOptionBase() {}



const std::string & SectionAndOptionBase::Name() const {
	return d_name;
}



const std::string & SectionAndOptionBase::Comment() const {
	return d_comment;
}


bool SectionAndOptionBase::IsUnknown() const {
	return d_isUnknown;
}



void SectionAndOptionBase::SetIsUnknown(bool isUnknown){
	d_isUnknown = isUnknown;
}



void SectionAndOptionBase::SetComment(const std::string & comment){
	d_comment = comment;
}



char options::SectionAndOptionBase::Short() const{
	return d_short;
}


void SectionAndOptionBase::SetNameFromString(const std::string & name) {
	std::string::size_type pos = name.find_first_of(',');
	if( pos != std::string::npos && pos != name.size() - 1){
		d_short = name.at(pos+1);
		d_name = name.substr(0,pos);
	} else {
		d_short = 0;
		d_name = name;
	}

	//forbid to use '.' in names ...
	pos = d_name.find_first_of('.');
	if( pos != std::string::npos ) {
		d_name.erase(pos);
	}
}



OptionBase::OptionBase(const std::string & name,
                       const std::string & comment)
	: SectionAndOptionBase(name,comment)
	, d_isDefault(true)
	, d_isUnknown(false){
}


OptionBase::~OptionBase(){}





bool OptionBase::IsDefault() const {
	return d_isDefault;
}


void OptionBase::SetIsDefault(bool def) {
	d_isDefault = def;
}





const std::string Option<bool>::s_true("true");
const std::string Option<bool>::s_false("false");

Section::Section(const std::string & name,
                 const std::string & comment)
	: SectionAndOptionBase(name,comment)
	, d_subSections()
	, d_options()
	, d_subSectionCreator(new SubSectionCreator()){
}



options::Section::~Section()
{
}

void options::Section::SetSubSectionCreator(SubSectionCreator::Ptr ssc){
	d_subSectionCreator = ssc;
}

const Section::List & Section::SubSections() const{
	return d_subSections;
}



Section::List & Section::SubSections(){
	return d_subSections;
}



const Section::Ptr Section::SubSection(const std::string & name) const{
	List::const_iterator fi = d_subSections.find(name);

	if ( fi == d_subSections.end() ){
		return Ptr();
	}

	return fi->second;
}



Section::Ptr Section::SubSection(const std::string & name){
	List::iterator fi = d_subSections.find(name);

	if ( fi == d_subSections.end() ){
		return Ptr();
	}

	return fi->second;
}



Section::Ptr Section::AddSubSection(const std::string & name, const
                                    std::string & comment){
	List::iterator fi = d_subSections.find(name);

	if(fi != d_subSections.end()){
		return fi->second;
	}

	Section::Ptr newSection = d_subSectionCreator->Create(name,comment);

	d_subSections.insert(std::make_pair(name,newSection));
	return newSection;
}

Section::SubSectionCreator::~SubSectionCreator(){}

Section::Ptr Section::SubSectionCreator::Create(const std::string & name,
                                                const std::string & comment){
	return Section::Ptr(new Section(name,comment));
}
const Section::ListOfOptions & Section::Options() const{
	return d_options;
}



Section::ListOfOptions & Section::Options(){
	return d_options;
}



const OptionBase::Ptr Section::GetOption(const std::string & name) const{
	ListOfOptions::const_iterator fi = d_options.find(name);

	if ( fi == d_options.end() ){
		return OptionBase::Ptr();
	}
	return fi->second;
}



OptionBase::Ptr Section::GetOption(const std::string & name){
	ListOfOptions::iterator fi = d_options.find(name);

	if ( fi == d_options.end() ){
		return OptionBase::Ptr();
	}

	return fi->second;
}



//Section::Ptr Section::DeepCopy(const std::string & newName) const{
//}
//
//
//
//Section::Ptr Section::DeepCopy() const{
//}


