/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file Parsers.cpp
 *
 *  \date Dec 1, 2011
 *  \author Alexandre Tuleu
 */

#include "Parsers.h"
#include <fstream>

using namespace options;


ConfigFileParser::ConfigFileParser()
	: d_comments()
	, d_currentSection()
	, d_rootSection()
	, d_currentSectionName(){
}
void ConfigFileParser::Parse(const std::string & filename, Section::Ptr root){
	std::ifstream file(filename.c_str());

	if(!file){
		return;
	}

	Parse(file,root);
}



void ConfigFileParser::Parse(std::istream & stream, Section::Ptr root){

	d_currentSection = root;
	d_rootSection = root;
	std::string line;
	ClearCommentStack();

	while(std::getline(stream,line)){
		std::string::size_type index(0);
		while(index < line.size() && std::isspace(line.at(index))){
			++index;
		}
		line.erase(0,index);

		if(line.size() == 0){
			continue;
		}

		if(line.at(0) == '#'){
			if(line.size() > 1 && line.at(1) == '#') {
				std::string c(line.substr(2));
				StripWhitespace(c);
				AppendNewLineToComment(c);
				continue;
			}
			//a single comment break the comment, so we push the stack
			PushNewCommentToStack();
			continue;
		}

		if(line.at(0) == '[') {

			index = line.find_first_of("]");

			if(index == std::string::npos){
				//to do exceptions ?
				continue;
			}

			//to do check that there is not a problem after ?
			line.erase(index);
			line.erase(0,1);

			line.erase(std::remove_if(line.begin(),line.end(),isspace),
			           line.end());

			//now select the current Suboption

			d_currentSection = GetSectionOrCreate(line);

			SetCommentFromStack(*d_currentSection);

		}

		index = line.find_first_of("=");

		if(index != std::string::npos) {
			std::string optionName(line.substr(0,index));

			optionName.erase(std::remove_if(optionName.begin(), optionName.end(), isspace),
			                 optionName.end());

			std::string value(line.substr(index + 1));

			index = value.find_first_of("#");
			if(index != std::string::npos){
				value.erase(index);
			}

			OptionBase::Ptr opt = GetOptionOrCreate(optionName);
			StripWhitespace(value);
			opt->Parse(value);


			SetCommentFromStack(*opt);
		}
	}
}



OptionBase::Ptr ConfigFileParser::GetOptionOrCreate(const std::string & name){
	OptionBase::Ptr option = d_currentSection->GetOption(name);
	if(option){
		return option;
	}
	return d_currentSection->AddOption<std::string>(name,"","");
}



Section::Ptr ConfigFileParser::GetSectionOrCreate(const std::string & name){
	std::istringstream iss(name);
	std::string sectionName;
	Section::Ptr section = d_rootSection;
	Section::Ptr subsection;

	while(std::getline(iss,sectionName,'.')){
		subsection = section->SubSection(sectionName);
		if(!subsection){
			subsection = section->AddSubSection(sectionName,"");
		}
		section = subsection;
	}

	return section;
}
void ConfigFileParser::SplitCommentInLines(std::string  comment,
                                           unsigned int maxLineLength,
                                           std::vector<std::string> & lines){


	lines.clear();

	std::string::size_type index;

	while(comment.size() > maxLineLength){

		index = comment.find_last_of(' ',maxLineLength);

		if(index == std::string::npos) {
			lines.push_back(comment);
			break;
		}

		lines.push_back(comment.substr(0,index));
		comment.erase(0,index+1);
	}

	lines.push_back(comment);
}


std::ostream & ConfigFileParser::Print(std::ostream & out,const Section::Ptr & root){
	//first print options :
	std::vector<std::string> splittedComment;
	for(Section::ListOfOptions::const_iterator o = root->Options().begin();
	    o != root->Options().end();
	    ++o){
		//std::ostringstream os;
		if(!o->second->Comment().empty()){
			SplitCommentInLines(o->second->Comment(),77,splittedComment);
			for(std::vector<std::string>::const_iterator l = splittedComment.begin();
			    l != splittedComment.end();
			    ++l){
				out << "## " << *l << std::endl;
			}
		}

		out << o->second->Name() << " = " << o->second->ValueAsString() << " # "
		    << (o->second->IsDefault() ? "(default value)"
		                               : "(custom value)")
		    << std::endl
		    << std::endl;
	}

	std::string savedName(d_currentSectionName);

	if(!root->Name().empty()){
		d_currentSectionName.append(root->Name()).append(".");
	}

	for(Section::List::const_iterator s = root->SubSections().begin();
	    s != root->SubSections().end();
	    ++s){
		if(!root->Name().empty()){
			out << std::endl << std::endl;
		}
		if(!s->second->Comment().empty()){
			SplitCommentInLines(s->second->Comment(),77,splittedComment);
			for(std::vector<std::string>::const_iterator l = splittedComment.begin();
			    l != splittedComment.end();
			    ++l){
				out << "## " << *l << std::endl;
			}
		}
		out << "[" << d_currentSectionName
		    << s->second->Name() << "]"
		    << std::endl
		    << std::endl;


		Print(out,s->second);
	}

	d_currentSectionName = savedName;

	return out;
}


void ConfigFileParser::StripWhitespace(std::string & str){
	if(str.empty()){
		return;
	}
	std::string::iterator beg = str.begin();


	while(beg != str.end() && ::isspace(*beg)){
		++beg;
	}
	str.erase(str.begin(),beg);

	if(str.empty()){
		return;
	}
	std::string::iterator end = str.end()-1;//exist because unempty
	while(end != str.begin() && ::isspace(*end)){
		--end;
	}
	++end;//just after last good char

	str.erase(end,str.end());


}


void ConfigFileParser::PushNewCommentToStack(){
	d_comments.push_back(std::string());
}

void ConfigFileParser::ClearCommentStack(){
	d_comments.clear();
	PushNewCommentToStack();
}


void ConfigFileParser::SetCommentFromStack(SectionAndOptionBase & so){
	//look through the stack for the first non empty comment, discards others
	for(StackOfString::reverse_iterator c = d_comments.rbegin();
	    c != d_comments.rend();
	    ++c){
		if(!(c->empty())){
			so.SetComment(*c);
			ClearCommentStack();
			return;
		}
	}
}

void ConfigFileParser::AppendNewLineToComment(const std::string & l){
	if(d_comments.back().empty()){
		d_comments.back().append(l);
		return;
	}

	if(l.empty()){
		return;
	}

	if(!std::isspace(d_comments.back().at(d_comments.back().size() - 1)) &&
	   !std::isspace(l.at(0))){
		d_comments.back().append(" ");
	}
	d_comments.back().append(l);

}
